#ifndef _POLYNOME_HPP_
#define _POLYNOME_HPP_


#include <iostream>


template <typename T>
class polynome;

template <typename T>
std::ostream& operator<<(std::ostream &, polynome<T> const&);

template <typename T>
class polynome
{
private:
  T *coeff;
  int degre;
public:
  // Trois constructeurs
  polynome(T* tab, int taille); // A partir d'un tableau de coefficients
  polynome(T c=0, int deg=0); // Pour creer le monome cX^deg
  polynome(polynome const&); // Par copie

  ~polynome()// Le destructeur
  {delete []coeff;}

  // L'operateur de copie
  polynome& operator=(polynome const&);

  // L'operateur P -> -P
  polynome operator-() const;


  friend polynome operator+(polynome const&P, polynome const&Q)
  {
  polynome const &plus_grand_deg = (P.degre >= Q.degre) ? P : Q;
  polynome const &autre = (P.degre >= Q.degre) ? Q : P;
  // "plus_grand_deg" et "autre" sont des references respectivement vers
  // le polynome de plus grand et de plus petit degre parmi P et Q.

  polynome somme = plus_grand_deg; // Ici, on appelle le constructeur par copie. Le polynome somme a deja le bon degre.
  for (int i=0; i<=autre.degre; i++)
	somme.coeff[i] += autre.coeff[i]; // On ajout les coefficients du polynome de plus petit degre.

  return somme;
  }


  friend polynome operator-(polynome const &P, polynome const &Q)
  {
	return P + (-Q);
  }

  friend polynome operator*(polynome const &P, polynome const &Q)
  {
	// Si un des polynomes est nul, on renvoie le polynome nul.
	if ( (P.degre == 0 && P.coeff[0]==0) || (Q.degre == 0 && Q.coeff[0]==0) )
	  return polynome(T(0),0);
	
	polynome prod (T(0), P.degre + Q.degre + 1);// On cree un polynome (nul) de bon degre.
	for (int i=0; i<=P.degre; i++)
	  {
		for (int j=0; j<=Q.degre; j++)
		  {
			prod.coeff[i+j] += P.coeff[i] * Q.coeff[j];
		  }  
	  }
	return prod;
  }
  
  
  polynome& operator+=(polynome const&);
  polynome& operator-=(polynome const&);
  polynome& operator*=(polynome const&);

  friend std::ostream& operator<< <T>(std::ostream &, polynome const&);

  polynome operator()(polynome const&) const; // La fonction Q -> P(Q)
  T operator()(T const&) const; // La fonction x -> P(x)
};





/****** Definitions des fonctions membre et des fonctions amies ******/

template <typename T>
polynome<T>::polynome(T* tab, int taille)
{
  coeff = new T[taille];
  degre = taille-1;
  for (int i=0; i<taille; i++)
	coeff[i] = tab[i];
}


template <typename T>
polynome<T>::polynome(T c, int deg)
{
  coeff = new T[deg+1];
  degre = deg;
  for (int i=0; i<degre; i++)
	coeff[i] = 0;
  coeff[degre] = c;
}

template <typename T>
polynome<T>::polynome(polynome<T> const &P)
{
  degre = P.degre;
  coeff = new T[degre+1];
  for (int i=0; i<=degre; i++)
	coeff[i] = P.coeff[i];
}

template <typename T>
polynome<T>& polynome<T>::operator=(polynome<T> const &P)
{
  if (this != &P)
	{
	  degre = P.degre;
	  coeff = new T[degre+1];
	  for (int i=0; i<=degre; i++)
		coeff[i] = P.coeff[i];
	}
  return *this;
}



template <typename T>
polynome<T> polynome<T>::operator-() const
{
  polynome P(P.coeff, P.degre+1);
  for (int i=0; i<=degre; i++)
	P.coeff[i] *= -1;
  return P;
}


template <typename T>
polynome<T>& polynome<T>::operator+=(polynome<T> const &P)
{
  *this = *this + P;
  return *this;
}

template <typename T>
polynome<T>& polynome<T>::operator-=(polynome<T> const &P)
{
  *this = *this - P;
  return *this;
}

template <typename T>
polynome<T>& polynome<T>::operator*=(polynome<T> const &P)
{
  *this = *this * P;
  return *this;
}

template <typename T>
std::ostream& operator<<(std::ostream &o, polynome<T> const &P)
{
  bool premier_coeff = false; // Vaut "faux", tant que l'on a pas affiche le premier coefficient non nul.
  for (int i=0; i<=P.degre; i++)
	{
	  if (P.coeff[i] != 0 && premier_coeff)
		o << " + ";
	  if (P.coeff[i] != 0)
		{
		  premier_coeff = true;
		  o << P.coeff[i];
		  if (i != 0)
			o << "X";
		  if (i>1)
			o << "^" << i;
		}
	}
  if (!premier_coeff)
	o << "0";
  return o;
}




template <typename T>
polynome<T> polynome<T>::operator()(polynome<T> const &Q) const
{
  polynome<T> composee=0;
  for (int i=degre; i>=0; i--)
	{
	  composee = composee * Q;
	  composee += coeff[i];
	}
  
  return composee;  
}


template <typename T>
T polynome<T>::operator()(T const &x) const
{
  T valeur=0;
  for (int i=degre; i>=0; i--)
	{
	  valeur = valeur * x;
	  valeur += coeff[i];
	}
  
  return valeur;  
}

  
  



#endif
